@@ -1,6 +1,3 @@
-import sys
-sys.path.append("../..")
-
 import logging
 logging.basicConfig(level = logging.INFO)
 
