@@ -1,3 +1,5 @@
+from __future__ import print_function, unicode_literals
+# -*- coding: utf8 -*-
 """
 protlib builds on the struct and SocketServer modules in the standard
 library to make it easy to implement binary network protocols. It provides 
@@ -15,14 +17,20 @@ import traceback
 from copy import deepcopy
 from select import select
 from warnings import warn
-from types import NoneType
+from StringIO import StringIO
 from time import mktime, time
 from datetime import datetime
-from StringIO import StringIO
 from logging.handlers import RotatingFileHandler
 from logging import getLogger, Formatter, NOTSET, DEBUG, INFO, WARNING, ERROR, CRITICAL
 from SocketServer import TCPServer, UDPServer, StreamRequestHandler, DatagramRequestHandler
 
+StringTypes = (type(b""), type(""))
+if sys.version_info < (3, 0):
+    BytesIO = StringIO
+    bytes = lambda x: str(bytearray(x))
+else:
+    from io import BytesIO
+
 __all__ = ["CError", "CWarning",
            "CType", "CStruct", "CStructType", "CArray",
            "CChar", "CUChar", "CShort", "CUShort", "CInt", "CUInt", "CLong", "CULong", "CFloat", "CDouble", "CString", "CUnicode",
@@ -36,17 +44,26 @@ class CError(ValueError):
 class CWarning(UserWarning):
     """the only warning class directly used by protlib (except DeprecationWarning)"""
 
-__version_info__ = (1, 3, 0, "final", 0)
+__version_info__ = (1, 4, 0, "final", 0)
 __version__ = "{0}.{1}.{2}".format(*__version_info__)
 
-BYTE_ORDER = "!"
-AUTOSIZED = "AUTO SIZING"
+BYTE_ORDER = b"!"
+AUTOSIZED = "AUTOSIZED"
+
+def _to_bytes(x):
+    s = x if isinstance(x, type(b"")) else x.__str__()
+    return s.encode("utf8") if isinstance(s, unicode) else s
 
 def _get_default(val):
     return val() if hasattr(val, "__call__") else deepcopy(val)
 
 def _fileize(x):
-    return StringIO(x) if isinstance(x, basestring) else x
+    if isinstance(x, type("")):
+        return StringIO(x)
+    elif isinstance(x, type(b"")):
+        return BytesIO(x)
+    else:
+        return x
 
 def _no_codec(name):
     try:
@@ -56,23 +73,25 @@ def _no_codec(name):
 
 def _is_open(sock):
     try:
-        return isinstance(sock.fileno(), int)
+        return isinstance(sock.fileno(), int) and sock.fileno() >= 0
     except:
         return False
 
 def _inherit_docstrings(klass):
-    for name,field in klass.__dict__.items():
+    for name,field in klass.__dict__.iteritems():
         parent = getattr(klass.__bases__[0], name, None)
         if hasattr(field, "__call__") and not field.__doc__ and parent:
             field.__doc__ = parent.__doc__
     return klass
 
 def _read_until_null(f):
-    s = ""
+    s = b""
     f = _fileize(f)
     while True:
         c = f.read(1)
-        if c in ["","\0"]:
+        if not c:
+            raise CError("end of file reached with no null byte found")
+        elif c == b"\0":
             break
         else:
             s += c      # shlemiel the painter
@@ -81,7 +100,7 @@ def _read_until_null(f):
 class _chained:
     def __init__(self, s, f):
         self.s, self.f = s, f
-        self.has_read = ""
+        self.has_read = b""
     
     def read(self, n):
         pos = len(self.has_read)
@@ -116,9 +135,6 @@ class CType(object):
         """
         CTypes have five optional keyword arguments:
         
-        length -- required for CString, CUnicode, and CArray,
-                  invalid for everything else
-        
         always -- Use this to set a constant value for a field. You won't
                   need to specify this value, and a warning will be raised
                   if the value ever differs from this parameter.
@@ -128,14 +144,21 @@ class CType(object):
                    serialized. Also, unlike an always parameter, a default 
                    parameter may be either a value or a callable object.
         
+        length -- required for CString, CUnicode, and CArray,
+                  invalid for everything else
+        
         encoding -- required for CUnicode, invalid for everything else
         
         enc_errors -- Only valid (but not required) for CUnicode, see the
                       unicode "errors" parameter for details.  This defaults
                       to "strict" but may be set to "ignore", "replace", etc
+        
+        full_string -- Only valid (but not required) for CString, set this
+                       to True to prevent parsed string values frome being
+                       truncated at the first null byte.
         """
-        self.always = self.default = self.length = self.encoding = self.enc_errors = None
-        extra = [name for name,val in settings.items() if not hasattr(self, name)]
+        self.always = self.default = self.length = self.encoding = self.enc_errors = self.full_string = None
+        extra = [name for name,val in settings.iteritems() if not hasattr(self, name)]
         if extra:
             warn("{0} settings do not include {1}".format(self.__class__.__name__, ", ".join(extra)), CWarning)
         
@@ -145,10 +168,12 @@ class CType(object):
         if self.length is not None and not isinstance(self, (CString, CUnicode, CArray)):
             warn("length has no meaning for {0} objects".format(self.__class__.__name__), CWarning)
         if isinstance(self, (CString, CUnicode, CArray)):
-            if not isinstance(self.length, (int, basestring)):
+            if not isinstance(self.length, int) and not isinstance(self.length, StringTypes):
                 raise CError("{0} objects require a string or integer length attribute".format(self.__class__.__name__))
             elif isinstance(self.length, int) and self.length < 0:
                 raise CError("length integer value must be positive")
+        if "full_string" in settings and not isinstance(self, CString):
+            raise warn("full_string parameter has no meaning for {0} objects".format(self.__class__.__name__), CWarning)
         
         if isinstance(self, CUnicode):
             self.enc_errors = "strict" if self.enc_errors is None else self.enc_errors
@@ -178,49 +203,83 @@ class CType(object):
         by extracting the actual value from the given cstruct instance.  This
         method raises an error if the computed length is negative.
         """
-        if isinstance(self.length, (int, NoneType)):
+        if isinstance(self.length, (int, type(None))):
             return self.length
         
         if cstruct is None:
             raise CError("cstruct not provided to resolve variable-length field with length attribute {0!r}".format(self.length))
         
         if self.length is AUTOSIZED:
-            field_name = dict((v,k) for k,v in cstruct.__dict__.items() if getattr(v, "__hash__", None))[self]
+            field_name = dict((v,k) for k,v in cstruct.__class__.__dict__.iteritems() if getattr(v, "__hash__", None))[self]
             field_val = getattr(cstruct, field_name)
-            if not isinstance(field_val, basestring):
-                raise CError("{0}.{1} is set to a non-string value: {2!r}".format(cstruct.__class__.__name__, field_name, field_val))
-            return len(field_val)
+            try:
+                return len(self.serialize(field_val, cstruct))
+            except:
+                exc_class, exc, tb = sys.exc_info()
+                cerror = CError("{0}.{1} is set to a non-string value: {2!r}".format(cstruct.__class__.__name__, field_name, field_val))
+                raise CError, cerror, tb
+            return len(serialized)
         else:
             length = getattr(cstruct, self.length, None)
-            if isinstance(length, (CType, NoneType)):
+            if isinstance(length, (CType, type(None))):
                 raise CError("{0}.{1} not set and is needed for use as a length attribute".format(cstruct.__class__.__name__, self.length))
             elif not isinstance(length, int):
-                raise CError("{0}.{1} must be an integer for use as a length attribute".format(cstruct.__class__.__name__, self.length))
+                raise CError("{0}.{1} must be an integer for use as a length attribute and was actually {2}".format(cstruct.__class__.__name__, self.length, length.__class__.__name__))
             elif length < 0:
                 raise CError("length field {0}.{1} may not be negative".format(cstruct.__class__.__name__, self.length))
             return length
     
     def struct_format(self, cstruct=None):
         """the format string used to represent this CType in the struct module"""
-        return {
-            CChar:    "b",
-            CUChar:   "B",
-            CShort:   "h",
-            CUShort:  "H",
-            CInt:     "i",
-            CUInt:    "I",
-            CLong:    "q",
-            CULong:   "Q",
-            CFloat:   "f",
-            CDouble:  "d",
-            CString:  "{0}s".format( self.real_length(cstruct) ),
-            CUnicode: "{0}s".format( self.real_length(cstruct) )
-        }[self.__class__]
+        formats = {
+            CChar:    b"b",
+            CUChar:   b"B",
+            CShort:   b"h",
+            CUShort:  b"H",
+            CInt:     b"i",
+            CUInt:    b"I",
+            CLong:    b"q",
+            CULong:   b"Q",
+            CFloat:   b"f",
+            CDouble:  b"d",
+            CString:  _to_bytes("{0}s".format(self.real_length(cstruct))),
+            CUnicode: _to_bytes("{0}s".format(self.real_length(cstruct)))
+        }
+        for ctype,format in formats.iteritems():
+            if isinstance(self, ctype):
+                return format
     
     def sizeof(self, cstruct=None):
         """the number of bytes of binary data needed to represent this CType"""
         return struct.calcsize(BYTE_ORDER + self.struct_format(cstruct))
     
+    def convert(self, x):
+        """
+        When a value is assigned to a struct field, this function
+        is converts the value appropriately, for example:
+        
+        >>> class Point(CStruct):
+        ...     x = CInt()
+        ...     y = CInt()
+        ... 
+        >>> Point(x = "5", y = 6.0)
+        Point(x=5, y=6)
+        
+        If you subclass one of the CType classes, you may need to
+        override this method.
+        """
+        if isinstance(self, CStructType):
+            return x
+        elif isinstance(self, CUnicode):
+            return x if isinstance(x, unicode) else unicode(_to_bytes(x), self.encoding, self.enc_errors)
+        elif isinstance(self, CArray):
+            return [self.ctype.convert(e) for e in x]
+        else:
+            for klass,converter in _converters.iteritems():
+                if isinstance(self, klass):
+                    return converter(x)
+            raise CError("no converter found for {0}".format(self.__class__.__name__))
+    
     def parse(self, f, cstruct=None):
         r"""
         Accepts either a string or a file and returns a Python object with 
@@ -229,7 +288,7 @@ class CType(object):
         
         CString parsing returns the string up to the first null byte, e.g.
         
-        CString(length=10).parse("foo\0barbaz") -> "foo"
+        CString(length=10).parse(b"foo\0barbaz") -> b"foo"
         """
         buf = _fileize(f).read( self.sizeof(cstruct) )
         if len(buf) < self.sizeof(cstruct):
@@ -247,15 +306,15 @@ class CType(object):
         
         Passing a too-short list to a CArray is okay if a default value
         was provided:
-            CArray(2, CChar(default=0)).serialize([1]) -> "\x01\x00"
+            CArray(2, CChar(default=0)).serialize([1]) -> b"\x01\0"
         
         Passing too much data to a CArray or CString will trigger a
         CWarning, e.g.
             CArray(2, CInt).serialize([5,6,7])
-            CString(length=3).serialize("Hello")
+            CString(length=3).serialize(b"Hello")
         
         Passing a too-short string to a CString is always okay:
-            CString(length=4).serialize("Hi") -> "Hi\x00\x00"
+            CString(length=4).serialize(b"Hi") -> b"Hi\0\0"
         """
         try:
             with warnings.catch_warnings():
@@ -283,11 +342,14 @@ class CString(CType):
         if self.length is AUTOSIZED:
             return _read_until_null(f)
         else:
-            return CType.parse(self, f, cstruct).split("\0")[0]
+            s = CType.parse(self, f, cstruct)
+            if not self.full_string:
+                s = s.split(b"\0")[0]
+            return s
     
     def serialize(self, val, cstruct=None):
         if self.length is AUTOSIZED:
-            return val.rstrip("\0") + "\0"
+            return val.split(b"\0", 1)[0] + b"\0"
         else:
             if len(val) > self.real_length(cstruct):
                 warn("CString has length {0} and was told to serialize a string of length {1}".format(self.real_length(cstruct), len(val)), CWarning)
@@ -297,7 +359,7 @@ class CString(CType):
 class CUnicode(CType):
     def parse(self, f, cstruct=None):
         if self.length is not AUTOSIZED:
-            return CType.parse(self, f, cstruct).split("\0")[0].decode(self.encoding, self.enc_errors)
+            return CType.parse(self, f, cstruct).split(b"\0")[0].decode(self.encoding, self.enc_errors)
         else:
             s = _read_until_null(f)
             try:
@@ -309,14 +371,14 @@ class CUnicode(CType):
     
     def serialize(self, val, cstruct=None):
         try:
-            encoded = _get_converter(self)(val).encode(self.encoding, self.enc_errors)
+            encoded = self.convert(val).encode(self.encoding, self.enc_errors)
         except:
             exc_class, exc, tb = sys.exc_info()
             cerror = CError("unicode error serializing {0!r}: {1}".format(val, exc))
             raise CError, cerror, tb
         
         if self.length is AUTOSIZED:
-            return encoded + "\0"
+            return encoded + b"\0"
         else:
             if len(encoded) > self.real_length(cstruct):
                 warn("CUnicode value has length {0} and was told to serialize an encoded string of length {1} {2!r}".format(self.real_length(cstruct), len(encoded), encoded), CWarning)
@@ -359,10 +421,10 @@ class CArray(CType):
         CType.__init__(self, length=length, **params)
         
         for param,value in [("always",self.always), ("default",self.default)]:
-            if value is not None and not isinstance(self.length, basestring):
+            if value is not None and not isinstance(self.length, StringTypes):
                 try:
                     value = _get_default(value) if param == "default" else value
-                    self.serialize( _get_converter(self)(value) )
+                    self.serialize( self.convert(value) )
                 except:
                     exc_class, exc, tb = sys.exc_info()
                     cerror = CError("{0!r} is not a valid {1} CArray value: {2}".format(value, param, exc))
@@ -389,8 +451,7 @@ class CArray(CType):
             if len(xs) < length:
                 raise CError("CArray has length {0} and was only given {1} elements".format(length, len(xs)))
         
-        xs = _get_converter(self)(xs)
-        return "".join(self.ctype.serialize(x, cstruct) for x in xs)
+        return b"".join(self.ctype.serialize(x, cstruct) for x in self.convert(xs))
 
 @_inherit_docstrings
 class CStructType(CType):
@@ -422,7 +483,7 @@ class CStructType(CType):
         CType.__init__(self, **params)
     
     def struct_format(self, cstruct=None):
-        return "".join(ctype.struct_format(cstruct) for name,ctype in self.subclass.get_fields())
+        return b"".join(ctype.struct_format(cstruct) for name,ctype in self.subclass.get_fields())
     
     def parse(self, f, cstruct=None):
         f = _fileize(f)
@@ -433,7 +494,7 @@ class CStructType(CType):
         return inst
     
     def serialize(self, inst, cstruct=None):
-        serialized = ""
+        serialized = b""
         for name,ctype in self.subclass.get_fields():
             val = inst.__dict__.get(name)
             if val is None:
@@ -445,7 +506,7 @@ class CStructType(CType):
                             if maybe is not None:
                                 val = []
                 elif isinstance(ctype, (CString, CUnicode)) and ctype.real_length(inst) == 0:
-                    val = ""
+                    val = b""
             if val is None:
                 raise CError(name + " not set")
             serialized += ctype.serialize(val, cstruct=inst)
@@ -469,26 +530,26 @@ class CStruct(CType):
         if not fields:
             raise CError("{0} struct contains no CType fields".format(self.__class__.__name__))
         
-        field_names = zip(*fields)[0]
+        field_names = list(zip(*fields))[0]
         for i,arg in enumerate(args):
             name = field_names[i]
             if name in values and values[name] != arg:
                 raise CError("{0} was given a value of {1!r} as a positional argument and {2!r} as a keyword argument".format(name, arg, values[name]))
             values[name] = arg
         
-        non_fields = [name for name,value in values.items() if name not in field_names]
+        non_fields = [name for name,value in values.iteritems() if name not in field_names]
         if non_fields:
             warn("{0} fields ({1}) do not include {2}".format(self.__class__.__name__, ", ".join(field_names), ", ".join(non_fields)), CWarning)
         
         for name,ctype in fields:
             if ctype.maybe is not None:
                 setattr(self, name, _get_default(ctype.maybe))
-            elif isinstance(ctype, (CString, CUnicode)) and isinstance(ctype.length, basestring) \
+            elif isinstance(ctype, (CString, CUnicode)) and isinstance(ctype.length, StringTypes) \
                     and ctype.length is not AUTOSIZED \
                     and not isinstance(getattr(self, ctype.length), CType) \
                     and ctype.real_length(self) == 0:
-                setattr(self, name, "")
-            elif isinstance(ctype, CArray) and isinstance(ctype.length, basestring) and ctype.ctype.maybe is not None \
+                setattr(self, name, b"")
+            elif isinstance(ctype, CArray) and isinstance(ctype.length, StringTypes) and ctype.ctype.maybe is not None \
                     and not isinstance(getattr(self, ctype.length), CType):
                 setattr(self, name, [_get_default(ctype.ctype.maybe) for i in xrange(ctype.real_length(self))])
             
@@ -510,12 +571,12 @@ class CStruct(CType):
             raise CError("CStruct classmethods may only be called on subclasses of CStruct")
         
         if "_fields" not in cls.__dict__:   # avoid hasattr because of subclasses
-            uninstantiated = [ctype for name,ctype in cls.__dict__.items()
+            uninstantiated = [ctype for name,ctype in cls.__dict__.iteritems()
                                     if type(ctype) is type and issubclass(ctype,CType)]
             if uninstantiated:
                 raise CError("Use {0}{2} instead of {0} when declaring a field in your {1} struct".format(uninstantiated[0].__name__, cls.__name__, ".get_type()" if issubclass(uninstantiated[0],CStruct) else "()"))
             
-            directly = [cstruct for name,cstruct in cls.__dict__.items() if isinstance(cstruct, CStruct)]
+            directly = [cstruct for name,cstruct in cls.__dict__.iteritems() if isinstance(cstruct, CStruct)]
             if directly:
                 raise CError("Use {0}.get_type() instead of {0}() when declaring a field in your {1} struct".format(directly[0].__class__.__name__, cls.__name__))
             
@@ -527,7 +588,7 @@ class CStruct(CType):
                     break
             
             if top is cls:
-                fields = [[name,ctype] for name,ctype in cls.__dict__.items() if isinstance(ctype, CType)]
+                fields = [[name,ctype] for name,ctype in cls.__dict__.iteritems() if isinstance(ctype, CType)]
                 fields.sort(key = lambda pair: CType.instances.index(pair[1]))
                 
                 positions = [(CType.instances.index(ctype),name,ctype) for name,ctype in fields]
@@ -542,13 +603,13 @@ class CStruct(CType):
                     if not isinstance(final, CType):
                         raise CError("{0} field overridden by non-CType {1!r}".format(pair[0], final))
                     elif type(final.length) is not type(pair[1].length) \
-                            or not isinstance(final.length, basestring) and final.sizeof() != pair[1].sizeof():
+                            or not isinstance(final.length, StringTypes) and final.sizeof() != pair[1].sizeof():
                         raise CError("{0[0]} field of type {0[1]} was overridden by differently-sized type {1}".format(pair, final.__class__.__name__))
                     pair[1] = final
             
             names = [name for name,ctype in fields]
             for (i,(name,ctype)) in enumerate(fields):
-                if isinstance(ctype.length, basestring) and ctype.length in names[i+1:]:      # shlemeil the painter
+                if isinstance(ctype.length, StringTypes) and ctype.length in names[i+1:]:      # shlemeil the painter
                     raise CError("{0}.{1} is the length field for {0}.{2} but appears after it in the struct".format(cls.__name__, ctype.length, name))
             
             cls._fields = fields
@@ -607,20 +668,20 @@ class CStruct(CType):
         """
         return self.get_type(cached=True).serialize(self)
     
-    __str__ = serialize
-    
     def __repr__(self):
         """
         Returns a literal representation of this struct.  This may be
         copy/pasted into a Python file or interpreter as valid code.  For
         example:
         
-        repr(Point(x=5, y=6)) == "Point(x=5, y=6)"
+        repr(Point(x=5, y=6)) == str(Point(x=5, y=6)) == "Point(x=5, y=6)"
         """
         params = ["{0}={1!r}".format(name, getattr(self,name))
                   for name,ctype in self.get_fields() if name in self.__dict__]
         return "{0}({1})".format(self.__class__.__name__, ", ".join(params))
     
+    __str__ = __repr__
+    
     @property
     def hashable(self):
         if not hasattr(self, "_hashable"):
@@ -655,7 +716,7 @@ class CStruct(CType):
                 raise CError("{0} assigned to the {1} field {2}.{3}".format(value.__class__.__name__, field.subclass.__name__, self.__class__.__name__, name))
         elif isinstance(field, CType):
             try:
-                value = _get_converter(field)(value)
+                value = field.convert(value)
             except:
                 exc_class, exc, tb = sys.exc_info()
                 cerror = CError("Conversion error: you provided the {0} value {1!r} to the {2} field {3}.{4}: {5}".format(value.__class__.__name__, value, field.__class__.__name__, self.__class__.__name__, name, exc))
@@ -676,9 +737,9 @@ class CStruct(CType):
 _converters = {
     CFloat:   float,
     CDouble:  float,
-    CString:  str,
-    CChar:    lambda c: ord(c) if isinstance(c, basestring) else int(c),
-    CUChar:   lambda c: ord(c) if isinstance(c, basestring) else int(c)
+    CString:  _to_bytes,
+    CChar:    lambda c: ord(c) if isinstance(c, StringTypes) else int(c),
+    CUChar:   lambda c: ord(c) if isinstance(c, StringTypes) else int(c)
 }
 def _to_int(x):
     if isinstance(x, float) and x != int(x):
@@ -686,24 +747,11 @@ def _to_int(x):
     return int(x)
 _converters.update((ctype, _to_int) for ctype in [CShort, CUShort, CInt, CUInt, CLong, CULong])
 
-def _get_converter(field):
-    if isinstance(field, CStructType):
-        return lambda cstruct: cstruct
-    elif isinstance(field, CUnicode):
-        return lambda x: x if isinstance(x, unicode) else unicode(str(x), field.encoding, field.enc_errors)
-    elif isinstance(field, CArray):
-        def converter(xs):
-            conv = _get_converter(field.ctype)
-            return [conv(x) for x in xs]
-        return converter
-    else:
-        return _converters.get(field.__class__)
-
 _formatter = Formatter("%(asctime)s: %(message)s")
 
 class _AlsoPrint(logging.Handler):
     def emit(self, record):
-        print( _formatter.format(record) )
+        print(_formatter.format(record))
 
 class _NullHandler(logging.Handler):
     def emit(self, record):
@@ -741,7 +789,8 @@ def hexdump(data):
     hexadecimal representation of the ordinal values of each character in
     the string passed as a parameter.
     """
-    hexed = [hex(ord(char))[2:].rjust(2,"0") for char in data]
+    pad = b"0" if isinstance(hex(0), type(b"")) else "0"
+    hexed = [hex(byte if isinstance(byte, int) else ord(byte))[2:].rjust(2, pad) for byte in data]
     lines = ["     0  1  2  3  4  5  6  7"]
     for i in xrange(0, len(hexed), 8):
         lines.append("%3i  " % i + " ".join(hexed[i:i+8]))
@@ -777,17 +826,18 @@ class Parser(object):
                  it was instantiated.  Structs defined afterwards will not
                  be detected.
         """
-        
         self.logger = logger or Logger(rfh_level=CRITICAL+1)
         
-        if not module:
+        if not module and hasattr(sys, "_getframe"):
             globs = sys._getframe().f_back.f_globals
+        elif not module:
+            raise CError("module argument is required for non-CPython interpreters")
         else:
-            if isinstance(module, basestring):
+            if isinstance(module, str):
                 module = __import__(module)
             globs = module.__dict__
         
-        self.structs = [cstruct for name,cstruct in globs.items()
+        self.structs = [cstruct for name,cstruct in globs.iteritems()
                         if type(cstruct) is type and issubclass(cstruct,CStruct) and cstruct is not CStruct]
         self.codes = []
         for cstruct in self.structs:
@@ -798,14 +848,14 @@ class Parser(object):
         if not self.codes:
             raise CError("No structs which begin with constant values were defined in the module " + (module.__name__ if module else "where you instantiated Parser"))
         
-        bufs = zip(*self.codes)[0]
+        bufs = list(zip(*self.codes))[0]
         while bufs:
             matches = [b for b in bufs if bufs[0] == b[:len(bufs[0])]]
             if len(matches) > 1:
                 structs = ", ".join(cstruct.__name__ for buf,cstruct in self.codes if buf in matches)
                 warn("{0} structs exist which always begin with {1!r}: {2}".format(len(matches), bufs[0], structs), CWarning)
             bufs = [b for b in bufs if b not in matches]
-
+    
     def parse(self, f):
         """
         Accepts a string or file object and returns a string or CStruct
@@ -818,7 +868,7 @@ class Parser(object):
            then an error is logged and None is returned.
         """
         f = _fileize(f)
-        buf = ""
+        buf = b""
         for code,cstruct in self.codes:
             diff = len(code) - len(buf)
             if diff:
@@ -991,9 +1041,11 @@ class ProtHandler(Logger):
             warn("LOG_DIR field is deprecated and ignored, you should use the logging.handler classes to affect log locations and properties", DeprecationWarning)
         if hasattr(self, "HEX_LOGGING"):
             warn("HEX_LOGGING field is deprecated and ignored, hex dumps are now written to their own separate log", DeprecationWarning)
+        if not self.STRUCT_MOD and not hasattr(sys, "_getframe"):
+            raise CError("STRUCT_MOD must be defined for non-CPython interpreters")
         
         Logger.__init__(self, prefix = self.LOG_PREFIX, also_print = self.LOG_TO_SCREEN,
-                              rfh_level = INFO if isinstance(server, (LoggingTCPServer, LoggingUDPServer, NoneType)) else 1+CRITICAL)
+                              rfh_level = INFO if isinstance(server, (LoggingTCPServer, LoggingUDPServer, type(None))) else 1+CRITICAL)
         self.__parser = Parser(logger=self, module=self.STRUCT_MOD or self.__class__.__module__)
         
         for buf,cstruct in self.__parser.codes:
@@ -1017,7 +1069,7 @@ class ProtHandler(Logger):
         and returns the result.  If no handler method exists for a cstruct,
         an error is logged.
         """
-        if isinstance(data, basestring):
+        if isinstance(data, StringTypes):
             return self.raw_data(data)
         
         codename = underscorize(data.__class__.__name__)
@@ -1089,10 +1141,10 @@ class TCPHandler(ProtHandler, StreamRequestHandler):
         This method returns quickly if this handler's server's socket has
         been closed.
         """
-        buf = ""
+        buf = b""
         timeout = self.timeout or self.server.timeout
-        wait = min(1, timeout or 1)
         until = timeout and (time() + timeout)
+        wait = min(1, timeout or 1)
         while _is_open(self.server.socket) and (not until or time() < until):
             r,_,_ = select([self.request], [], [], wait)
             if r:
