@@ -1,21 +1,26 @@
 # -*- coding: latin-1 -*-
-import protlib
-from protlib import *
-
+from __future__ import unicode_literals, division
 import os
 import sys
+import json
 import math
 import atexit
 import socket
 import logging
+import warnings
 from glob import glob
 from time import sleep, time
 from threading import Thread
-from StringIO import StringIO
-
 from unittest import TestCase, main
 
-import warnings
+import protlib
+from protlib import *
+
+if sys.version_info < (3, 0):
+    from StringIO import StringIO as BytesIO
+else:
+    from io import BytesIO
+
 warnings.simplefilter("error", CWarning)
 
 def assert_works_without_warnings(func, *args, **kwargs):
@@ -25,7 +30,11 @@ def assert_works_without_warnings(func, *args, **kwargs):
         protlib.__warningregistry__.clear()
 
 def delete_logs():
-    atexit._exithandlers[:] = []    # stop the logging module's exit handler
+    if hasattr(atexit, "_clear"):
+        atexit._clear()
+    elif hasattr(atexit, "_exithandlers"):
+        atexit._exithandlers[:] = []    # stop the logging module's exit handler
+    
     for suffix in Logger.LEVELS:
         filename = "unit_tests.{0}_log".format(suffix)
         if os.path.exists(filename):
@@ -38,8 +47,8 @@ class NamedPoint(CStruct):
     code = CShort(always = 0x1234)
     x    = CInt()
     y    = CInt()
-    name = CString(length=15, default="unnamed")
-NP_BUF = "\x124\x00\x00\x00\x05\x00\x00\x00\x06unnamed\x00\x00\x00\x00\x00\x00\x00\x00"
+    name = CString(length=15, default=b"unnamed")
+NP_BUF = b"\x124\0\0\0\x05\0\0\0\x06unnamed\0\0\0\0\0\0\0\0"
 
 class RenamedPoint(NamedPoint):
     code = CShort(always = 0x4321)
@@ -54,6 +63,36 @@ class PointGroup(CStruct):
     count  = CShort()
     points = CArray("count", NamedPoint)
 
+class Person(CStruct):
+    code = CShort(always = 9)
+    name = CString(length = AUTOSIZED)
+
+class EvenPaddedCString(CString):
+    def __init__(self, **params):
+        CString.__init__(self, length=AUTOSIZED, **params)
+    
+    def parse(self, f, cstruct=None):
+        s = CString.parse(self, f, cstruct)
+        if len(s) % 2 == 0:
+            f.read(1)
+        return s
+    
+    def serialize(self, s, cstruct=None):
+        s = CString.serialize(self, s, cstruct)
+        if len(s) % 2:
+            s += b"\0"
+        return s
+
+class JsonCString(CString):
+    def parse(self, f, cstruct=None):
+        return json.loads(CString.parse(self, f, cstruct).decode("utf8"))
+    
+    def serialize(self, s, cstruct=None):
+        return CString.serialize(self, json.dumps(s).encode("utf8"), cstruct)
+    
+    def convert(self, x):
+        return x
+
 SERVER_ADDR = ("127.0.0.1", 7357)
 CLIENT_ADDR = ("127.0.0.1", 5737)
 
@@ -62,7 +101,7 @@ class TestHandler(object):
         return RenamedPoint(x=np.x, y=np.y)
     
     def renamed_point(self, rp):
-        return "Hello World!\n"
+        return b"Hello World!\n"
     
     def point_group(self, pg):
         pg.count *= 2
@@ -87,10 +126,11 @@ class CTypeTests(TestCase):
         self.assertRaises(CError, CType)
         self.assertRaises(CWarning, CInt, length=5)
         self.assertRaises(CWarning, CInt, something=6)
+        self.assertRaises(CWarning, CInt, full_string=True)
         self.assertRaises(TypeError, CInt, 5)
         self.assertRaises(TypeError, CString, 5)
-        self.assertRaises(CError, CInt().parse, "")
-        self.assertRaises(CError, CArray(2,CInt).parse, "1234")
+        self.assertRaises(CError, CInt().parse, b"")
+        self.assertRaises(CError, CArray(2,CInt).parse, b"1234")
     
     def test_integer_boundaries(self):
         for signed,unsigned,exp in [(CChar,CUChar,8), (CShort,CUShort,16), (CInt,CUInt,32), (CLong,CULong,64)]:
@@ -122,17 +162,21 @@ class CTypeTests(TestCase):
     
     def test_valid_cstring(self):
         cs = CString(length = 20)
-        self.assertEqual(20, len(cs.serialize("Hello World!")))
-        self.assertEqual("Hello World!", cs.parse("Hello World!" + "\x00" * 8))
-        self.assertEqual("Hello World!", cs.parse("Hello World!\x001234567"))
+        self.assertEqual(20, len(cs.serialize(b"Hello World!")))
+        self.assertEqual(b"Hello World!", cs.parse(b"Hello World!" + b"\0" * 8))
+        self.assertEqual(b"Hello World!", cs.parse(b"Hello World!\x001234567"))
         
         cs = CString(length = 100)
-        self.assertEqual("Hello World!", cs.parse(cs.serialize("Hello World!")))
+        self.assertEqual(b"Hello World!", cs.parse(cs.serialize(b"Hello World!")))
+        
+        cs = CString(length=10, full_string=True)
+        self.assertEqual(b"Hello\0\0\0\0\0", cs.parse(cs.serialize(b"Hello")))
     
     def test_invalid_cstring(self):
         self.assertRaises(CError, CString)
+        self.assertRaises(CError, CString, length=-1)
         self.assertRaises(CError, CString, length=None)
-        self.assertRaises(CWarning, CString(length=2).serialize, "Hello")
+        self.assertRaises(CWarning, CString(length=2).serialize, b"Hello")
     
     def test_array_instantiation(self):
         class Point(CStruct):
@@ -146,7 +190,7 @@ class CTypeTests(TestCase):
     
     def test_array_packing(self):
         xs = CArray(2, CInt, default=[0,6])
-        buf = "\x00\x00\x00\x05\x00\x00\x00\x06"
+        buf = b"\0\0\0\x05\0\0\0\x06"
         self.assertEqual(xs.serialize([5,6]), buf)
         self.assertEqual(xs.serialize([5]), buf)
         self.assertEqual([5,6], xs.parse(buf))
@@ -223,7 +267,7 @@ class CTypeTests(TestCase):
         dup = NamedPoint(0x1234, 5, 6, x=5, y=6)
         parsed = NamedPoint.parse(buf)
         evaled = eval( repr(np) )
-        from_file = NamedPoint.parse( StringIO(NP_BUF) )
+        from_file = NamedPoint.parse( BytesIO(NP_BUF) )
         
         self.assertEqual(buf, NP_BUF)
         self.assertEqual(np, pos)
@@ -251,9 +295,9 @@ class CTypeTests(TestCase):
         class Segment(CStruct):
             p1 = NamedPoint.get_type()
             p2 = NamedPoint.get_type()
-        Segment(p1=NamedPoint(x=5,y=6), p2=NamedPoint(x=11,y=42))
         Segment.parse(NP_BUF * 2)
-        Segment.parse(NP_BUF * 2 + "extra data in buffer")
+        Segment.parse(NP_BUF * 2 + b"extra data in buffer")
+        Segment(p1=NamedPoint(x=5,y=6), p2=NamedPoint(x=11,y=42))
     
     def test_struct_arrays(self):
         class Segment(CStruct):
@@ -275,7 +319,7 @@ class CTypeTests(TestCase):
         self.assertRaises(CError, NamedPoint, 0x1234, 5, x=6)
         self.assertRaises(CError, NamedPoint(x=5).serialize)
         self.assertRaises(CWarning, NamedPoint, code=0x4321, x=5, y=6)
-        self.assertRaises(CWarning, NamedPoint.parse, "!C\x00\x00\x00\x05\x00\x00\x00\x06unnamed\x00\x00\x00\x00\x00\x00\x00\x00")
+        self.assertRaises(CWarning, NamedPoint.parse, b"!C\0\0\0\x05\0\0\0\x06unnamed\0\0\0\0\0\0\0\0")
     
     def test_invalid_structs(self):
         self.assertRaises(CError, NamedPoint.parse, NP_BUF[:-1])
@@ -283,12 +327,12 @@ class CTypeTests(TestCase):
         class Point(CStruct):
             pass
         self.assertRaises(CError, Point)
-        self.assertRaises(CError, Point.parse, "\x00\x00\x00\x00")
+        self.assertRaises(CError, Point.parse, b"\0\0\0\0")
         
         class Point(CStruct):
             x = CInt
         self.assertRaises(CError, Point)
-        self.assertRaises(CError, Point.parse, "\x00\x00\x00\x00")
+        self.assertRaises(CError, Point.parse, b"\0\0\0\0")
         
         class Segment(CStruct):
             p1 = NamedPoint
@@ -314,7 +358,12 @@ class CTypeTests(TestCase):
         class Point(CStruct):
             x = y = CInt()
         self.assertRaises(CWarning, Point)
-        self.assertRaises(CWarning, Point.parse, "\x00" * 8)
+        self.assertRaises(CWarning, Point.parse, "\0" * 8)
+    
+    def test_repr_eval(self):
+        np = NamedPoint.parse(NP_BUF)
+        pg = PointGroup(count=2, points=[np, np])
+        self.assertEqual(pg, eval(repr(pg)))
     
     def test_valid_inheritance(self):
         class Origin(NamedPoint):
@@ -334,7 +383,7 @@ class CTypeTests(TestCase):
     
     def test_type_coercion(self):
         self.assertEqual(5, NamedPoint(x="5").x)
-        self.assertEqual("6", NamedPoint(name=6).name)
+        self.assertEqual(b"6", NamedPoint(name=6).name)
         
         class Letter(CStruct):
             c = CChar()
@@ -348,12 +397,12 @@ class CTypeTests(TestCase):
         class Point(CStruct):
             xy = CArray(2, CChar, default=[5,"\x06"])
         self.assertEqual(Point().xy, [5,6])
-        self.assertEqual(Point(xy=[1]).serialize(), "\x01\x06")
+        self.assertEqual(Point(xy=[1]).serialize(), b"\x01\x06")
         
         class Point(CStruct):
             xy = CArray(2, CChar, default=lambda: [5,"\x06"])
         self.assertEqual(Point().xy, [5,6])
-        self.assertEqual(Point(xy=[1]).serialize(), "\x01\x06")
+        self.assertEqual(Point(xy=[1]).serialize(), b"\x01\x06")
         
         self.assertRaises(CError, NamedPoint, x=5.6)
         self.assertRaises(CError, CArray, 2, CInt, always=[0, 1.2])
@@ -379,11 +428,11 @@ class CTypeTests(TestCase):
         CUnicode(length=5, encoding="utf8", enc_errors="strict")
         CUnicode(length=5, encoding="utf8", enc_errors="ignore")
         CUnicode(length=5, encoding="utf8", enc_errors="replace")
-        CUnicode(length=5, encoding="utf8", default=u"hello")
+        CUnicode(length=5, encoding="utf8", default="hello")
         CUnicode(length=5, encoding="utf8", default="hello")
     
     def test_unicode_integrity(self):
-        andre = u"andr"
+        andre = "andr"
         
         cu = CUnicode(length=6, encoding="latin-1")
         self.assertEqual(andre, cu.parse(cu.serialize(andre)))
@@ -397,20 +446,22 @@ class CTypeTests(TestCase):
         self.assertEqual(p.name, Person.parse(s).name)
     
     def test_unicode_coercion(self):
-        andre = u"andr"
+        andre = "andr"
         class Name(CStruct):
             first = CString(length=9)
             last = CUnicode(length=9, encoding="utf8")
         
-        self.assertRaises(CError, Name, first=andre)
-        name = Name(first=u"andre", last="giant")
-        self.assertTrue(isinstance(name.first, str))
+        if sys.version_info < (3, 0):
+            self.assertRaises(CError, Name, first=andre)
+        
+        name = Name(first="andre", last="giant")
+        self.assertTrue(isinstance(name.first, bytes))
         self.assertTrue(isinstance(name.last, unicode))
         
         name = Name(first=5, last=5)
         self.assertEqual(name.last, "5")
-        self.assertEqual(name.first, "5")
-        self.assertTrue(isinstance(name.first, str))
+        self.assertEqual(name.first, b"5")
+        self.assertTrue(isinstance(name.first, bytes))
         self.assertTrue(isinstance(name.last, unicode))
         
         s = CUnicode(length=6, encoding="utf8").serialize(andre)
@@ -419,25 +470,25 @@ class CTypeTests(TestCase):
     def test_encoding_errors(self):
         class Person(CStruct):
             name = CUnicode(length=9, encoding="utf8")
-        self.assertRaises(CError, Person, name="\x80")
-        self.assertRaises(CError, setattr, Person(), "name", "\x80")
+        self.assertRaises(CError, Person, name=b"\x80")
+        self.assertRaises(CError, setattr, Person(), "name", b"\x80")
         
         class Person(CStruct):
             name = CUnicode(length=9, encoding="utf8", enc_errors="ignore")
-        self.assertEqual("", Person(name="\x80").name)
+        self.assertEqual("", Person(name=b"\x80").name)
         p = Person()
-        p.name = "\x80"
+        p.name = b"\x80"
         self.assertEqual("", p.name)
         
         class Person(CStruct):
             name = CUnicode(length=9, encoding="utf8", enc_errors="replace")
-        self.assertEqual(u"\ufffd", Person(name="\x80").name)
+        self.assertEqual("\ufffd", Person(name=b"\x80").name)
         p = Person()
-        p.name = "\x80"
-        self.assertEqual(u"\ufffd", p.name)
+        p.name = b"\x80"
+        self.assertEqual("\ufffd", p.name)
     
     def test_length_problems(self):
-        self.assertRaises(CWarning, CUnicode(length=5, encoding="utf8").serialize, u"andr")
+        self.assertRaises(CWarning, CUnicode(length=5, encoding="utf8").serialize, "andr")
         
         class Person(CStruct):
             name = CString(length=5, default="marvin")
@@ -451,27 +502,27 @@ class CTypeTests(TestCase):
         assert_works_without_warnings(setattr, p, "name", "marvin")
         
         class Person(CStruct):
-            name = CUnicode(length=5, encoding="utf8", default=u"andr")
+            name = CUnicode(length=5, encoding="utf8", default="andr")
         self.assertRaises(CError, Person)
         assert_works_without_warnings(Person)
         
         class Person(CStruct):
             name = CUnicode(length=5, encoding="utf8")
         p = Person()
-        self.assertRaises(CError, setattr, p, "name", u"andr")
-        assert_works_without_warnings(setattr, p, "name", u"andr")
+        self.assertRaises(CError, setattr, p, "name", "andr")
+        assert_works_without_warnings(setattr, p, "name", "andr")
         
         class Name(CStruct):
             first = CString(length=5)
             last = CUnicode(length=5, encoding="utf8")
         self.assertRaises(CError, Name, first="marvin")
-        self.assertRaises(CError, Name, last=u"andr")
+        self.assertRaises(CError, Name, last="andr")
         assert_works_without_warnings(Name, first="marvin")
-        assert_works_without_warnings(Name, last=u"andr")
+        assert_works_without_warnings(Name, last="andr")
 
 class VarlengthTests(TestCase):
     def test_varlength_structs(self):
-        buf = "\xff\x00\x02" + NP_BUF + NP_BUF
+        buf = b"\xff\0\x02" + NP_BUF + NP_BUF
         pg = PointGroup(count=2, points=[NamedPoint(x=5,y=6), NamedPoint(x=5,y=6)])
         self.assertEqual(pg.serialize(), buf)
         self.assertEqual(pg, PointGroup.parse(buf))
@@ -480,12 +531,12 @@ class VarlengthTests(TestCase):
         class GeoStuff(CStruct):
             code    = CShort(always = 255)
             pg      = PointGroup.get_type()
-            message = CString(length=100, default="hw")
+            message = CString(length=100, default=b"hw")
         
-        buf = "\x00\xff\xff\x00\x02" + NP_BUF + NP_BUF + "hw" + "\x00" * 98
+        buf = b"\0\xff\xff\0\x02" + NP_BUF + NP_BUF + b"hw" + b"\0" * 98
         pg = PointGroup(count=2, points=[NamedPoint(x=5,y=6), NamedPoint(x=5,y=6)])
         gs = GeoStuff(pg = pg)
-        self.assertEqual(gs.message, "hw")
+        self.assertEqual(gs.message, b"hw")
         self.assertEqual(gs.serialize(), buf)
         self.assertEqual(gs, GeoStuff.parse(buf))
     
@@ -497,7 +548,7 @@ class VarlengthTests(TestCase):
             farewell = CString(length = "flen")
             version  = CChar(always = 9)
         
-        buf = "\x05hello\x07goodbye\x09"
+        buf = b"\x05hello\x07goodbye\x09"
         mess = Messages(greeting="hello", farewell="goodbye", glen=5, flen=7)
         self.assertEqual(mess.version, 9)
         self.assertEqual(mess.serialize(), buf)
@@ -510,9 +561,9 @@ class VarlengthTests(TestCase):
             
             @property
             def real_size(self):
-                return self.size / 2
+                return self.size // 2
         
-        buf = "\x00\x06\x01\x02\x03"
+        buf = b"\0\x06\x01\x02\x03"
         ann = AnnoyingSizing(size=6, arr=[1,2,3])
         self.assertEqual(ann.serialize(), buf)
         self.assertEqual(ann, AnnoyingSizing.parse(buf))
@@ -521,13 +572,13 @@ class VarlengthTests(TestCase):
         self.assertRaises(CError, AnnoyingSizing.parse, buf)
     
     def test_varlength_standalone(self):
-        mock = type("mock", (), {"foo": 15})
+        mock = type(type.__name__, (), {"foo": 15})
         
         cs = CString(length = "foo")
-        buf = cs.serialize("Hello World!", cstruct=mock)
+        buf = cs.serialize(b"Hello World!", cstruct=mock)
         s = cs.parse(buf, cstruct=mock)
         self.assertEqual(15, len(buf))
-        self.assertEqual(s, "Hello World!")
+        self.assertEqual(s, b"Hello World!")
         
         ca = CArray("foo", CChar(default=0))
         buf = ca.serialize([1,2,3], cstruct=mock)
@@ -536,12 +587,12 @@ class VarlengthTests(TestCase):
         self.assertEqual(xs, [1,2,3] + [0]*12)
     
     def test_varlength_attrs(self):
-        mock = type("mock", (), {"foo": 3})
+        mock = type(type.__name__, (), {"foo": 3})
         pg = PointGroup(count=1, points=[NamedPoint(x=5,y=6)])
-        self.assertEqual(28, PointGroup.sizeof(cstruct = pg))
-        self.assertEqual("Bh" + NamedPoint.struct_format(), PointGroup.struct_format(cstruct = pg))
+        self.assertEqual(28, PointGroup.sizeof(pg))
+        self.assertEqual(b"Bh" + NamedPoint.struct_format(), PointGroup.struct_format(cstruct = pg))
         self.assertEqual(12, CArray("foo", CInt).sizeof(cstruct = mock))
-        self.assertEqual("hhh", CArray("foo", CShort).struct_format(cstruct = mock))
+        self.assertEqual(b"hhh", CArray("foo", CShort).struct_format(cstruct = mock))
     
     def test_varlength_missing_length(self):
         self.assertRaises(CError, CString(length="foo").sizeof)
@@ -563,7 +614,7 @@ class VarlengthTests(TestCase):
         self.assertRaises(CError, VarLen, arr=[1,2])
         self.assertRaises(CError, VarLen, size=2, arr=[1])
         self.assertRaises(CError, VarLen, size=2, arr=[1,2,3])
-        self.assertRaises(CError, VarLen.parse, "\x02\x05")
+        self.assertRaises(CError, VarLen.parse, b"\x02\x05")
     
     def test_varlength_bad_order(self):
         class BadVarlength(CStruct):
@@ -572,14 +623,14 @@ class VarlengthTests(TestCase):
         
         self.assertRaises(CError, BadVarlength)
         self.assertRaises(CError, BadVarlength, size=2, arr=[5,6])
-        self.assertRaises(CError, BadVarlength.parse, "\x05\x06\x02")
+        self.assertRaises(CError, BadVarlength.parse, b"\x05\x06\x02")
     
     def test_negative_varlength(self):
         class NegVarLength(CStruct):
             size = CChar()
             arr  = CArray("size", CInt(default=0))
         self.assertRaises(CError, NegVarLength, -1)
-        self.assertRaises(CError, NegVarLength.parse, "\xFF")
+        self.assertRaises(CError, NegVarLength.parse, b"\xFF")
     
     def test_varlength_matrix(self):
         for outer_size in [2, "size"]:
@@ -587,7 +638,7 @@ class VarlengthTests(TestCase):
                 size   = CChar()
                 matrix = CArray(outer_size, CArray("size", CChar))
             
-            buf = "\x02\x05\x06\x07\x0b"
+            buf = b"\x02\x05\x06\x07\x0b"
             mat = Matrix(size=2, matrix=[[5,6],[7,11]])
             self.assertEqual(mat.serialize(), buf)
             self.assertEqual(mat, Matrix.parse(buf))
@@ -596,19 +647,19 @@ class VarlengthTests(TestCase):
         class DefaultEmpty(CStruct):
             size   = CChar(default = 0)
             string = CString(length = "size")
-        self.assertEqual("\x00", DefaultEmpty.parse("\x00").serialize())
-        self.assertEqual("", DefaultEmpty.parse("\x00").string)
-        self.assertEqual("", DefaultEmpty().string)
+        self.assertEqual(b"\0", DefaultEmpty.parse(b"\0").serialize())
+        self.assertEqual(b"", DefaultEmpty.parse(b"\0").string)
+        self.assertEqual(b"", DefaultEmpty().string)
         
         class DefaultEmpty(CStruct):
             size = CChar(default = 0)
             arr  = CArray("size", CInt)
-        self.assertEqual("\x00", DefaultEmpty.parse("\x00").serialize())
+        self.assertEqual(b"\0", DefaultEmpty.parse(b"\0").serialize())
         
         class DefaultConflict(CStruct):
             size = CChar(default = 2)
             arr = CArray("size", CInt, default=[])
-        self.assertRaises(CError, DefaultConflict.parse, "\x00")
+        self.assertRaises(CError, DefaultConflict.parse, b"\0")
         self.assertRaises(CError, DefaultConflict)
         
         class DerivedDefault(CStruct):
@@ -616,24 +667,47 @@ class VarlengthTests(TestCase):
             arr = CArray("size", CInt(default=5))
         self.assertEqual(DerivedDefault().arr, [5,5])
 
-class UnsizedTests(TestCase):
+class AutosizedTests(TestCase):
     def test_basic(self):
-        self.assertEqual("hello", CString(length=AUTOSIZED).parse("hello\0world"))
-        self.assertEqual("hello\0", CString(length=AUTOSIZED).serialize("hello"))
-        self.assertEqual("hello\0", CString(length=AUTOSIZED).serialize("hello\0\0\0"))
+        self.assertEqual(b"hello", CString(length=AUTOSIZED).parse(b"hello\0world"))
+        self.assertEqual(b"hello\0", CString(length=AUTOSIZED).serialize(b"hello"))
+        self.assertEqual(b"hello\0", CString(length=AUTOSIZED).serialize(b"hello\0\0\0"))
     
     def test_structs(self):
-        buf = "\0\x01asher\0\0\x04"
+        buf = b"\0\x01asher\0\0\x04"
         class Person(CStruct):
             code = CShort(always = 1)
             name = CString(length = AUTOSIZED)
             age = CShort()
         p = Person.parse(buf)
         self.assertEqual(p.code, 1)
-        self.assertEqual(p.name, "asher")
+        self.assertEqual(p.name, b"asher")
         self.assertEqual(p.age, 4)
         self.assertEqual(buf, p.serialize())
     
+    def test_parsing(self):
+        buf = Person(name="Eli").serialize()[:-1]
+        self.assertRaises(CError, CString(length=AUTOSIZED).parse, b"Eli")
+        self.assertRaises(CError, Person.parse, b"\0\x09Eli")
+    
+    def test_sizeof(self):
+        class Person(CStruct):
+            name = CString(length = AUTOSIZED)
+        self.assertRaises(CError, Person.sizeof)
+        self.assertRaises(CError, Person.sizeof, Person())
+        self.assertEqual(4, Person.sizeof(Person(name="Eli")))
+        
+        class Person(CStruct):
+            name = CString(length = AUTOSIZED)
+            age = CShort()
+        self.assertRaises(CError, Person.sizeof)
+        self.assertEqual(6, Person.sizeof(Person(name="Eli")))
+        
+        class Person(CStruct):
+            name = CString(length=AUTOSIZED, default="")
+            age = CShort()
+        self.assertEqual(3, Person.sizeof(Person()))
+    
     def test_nested(self):
         class Person(CStruct):
             code = CChar(always = 1)
@@ -644,18 +718,18 @@ class UnsizedTests(TestCase):
             name = CString(length = AUTOSIZED)
             people = CArray(2, Person.get_type(default=Person(name="", age=0)))
         
-        default_person = "\x01\0\0\0"
-        buf = "foo\0" + default_person * 2
+        default_person = b"\x01\0\0\0"
+        buf = b"foo\0" + default_person * 2
         self.assertEqual(buf, Department("foo").serialize())
         self.assertEqual(Department("foo"), Department.parse(buf))
         
-        buf = "foo\0\x01Eli\0\0\x1c\x01Asher\0\0\x04"
+        buf = b"foo\0\x01Eli\0\0\x1c\x01Asher\0\0\x04"
         dept = Department("foo", [Person(name="Eli", age=28), Person(name="Asher", age=4)])
         self.assertEqual(buf, dept.serialize())
         self.assertEqual(dept, Department.parse(buf))
     
     def test_unicode(self):
-        andre = u"andr"
+        andre = "andr"
         
         class Person(CStruct):
             code = CChar(always = 1)
@@ -664,27 +738,87 @@ class UnsizedTests(TestCase):
         
         class Department(CStruct):
             name = CUnicode(length=AUTOSIZED, encoding="utf8")
-            people = CArray(2, Person.get_type(default=Person(name=u"", age=0)))
+            people = CArray(2, Person.get_type(default=Person(name="", age=0)))
         
-        default_person = "\x01\0\0\0"
-        buf = "foo\0" + default_person * 2
+        default_person = b"\x01\0\0\0"
+        buf = b"foo\0" + default_person * 2
         self.assertEqual(buf, Department("foo").serialize())
-        self.assertEqual(Department("foo"), Department.parse(buf))
+        self.assertEqual(Department(b"foo"), Department.parse(buf))
         
-        buf = "foo\0\x01andr\xc3\xa9\0\0\x1c\x01Asher\0\0\x04"
-        dept = Department("foo", [Person(name=andre, age=28), Person(name="Asher", age=4)])
+        buf = b"foo\0\x01andr\xc3\xa9\0\0\x1c\x01Asher\0\0\x04"
+        dept = Department(b"foo", [Person(name=andre, age=28), Person(name=b"Asher", age=4)])
         self.assertEqual(buf, dept.serialize())
         self.assertEqual(dept, Department.parse(buf))
     
     def test_unicode_null_problem(self):
-        s = u"Hello World!".encode("utf-32")
-        self.assertTrue(s.count("\0"))
+        s = "Hello World!".encode("utf-32")
+        self.assertTrue(s.count(b"\0"))
         self.assertRaises(CError, CUnicode(length=AUTOSIZED, encoding="utf-32").parse, s)
         
         class Person(CStruct):
             name = CUnicode(length=AUTOSIZED, encoding="utf-32")
         self.assertRaises(CError, Person.parse, s)
+
+class SubclassedCTypeTests(TestCase):
+    def test_even_padded(self):
+        class Person(CStruct):
+            name = EvenPaddedCString()
+            age = CShort()
         
+        asher = Person(name="Asher", age=4)
+        buf = asher.serialize()
+        self.assertEqual(buf, b"Asher\0\0\x04")
+        self.assertEqual(asher, Person.parse(buf))
+        self.assertEqual(b"Asher", Person.parse(buf).name)
+        
+        liam = Person(name="Liam", age=1)
+        buf = liam.serialize()
+        self.assertEqual(buf, b"Liam\0\0\0\x01")
+        self.assertEqual(liam, Person.parse(buf))
+        self.assertEqual(b"Liam", Person.parse(buf).name)
+    
+    def test_json(self):
+        class Person(CStruct):
+            name = CString(length = AUTOSIZED)
+            data = JsonCString(length = AUTOSIZED)
+        
+        asher = Person(name="Asher", data={"age": 4})
+        buf = asher.serialize()
+        self.assertEqual(asher, Person.parse(buf))
+        self.assertEqual(buf, b'Asher\0{"age": 4}\0')
+        self.assertEqual({"age": 4}, Person.parse(buf).data)
+        
+        person = Person()
+        self.assertRaises(Exception, Person.parse, "Eli\0{\0")
+        self.assertRaises(Exception, setattr, person, "data", {"unserializable": protlib})
+    
+    def test_sizeof(self):
+        class Person(CStruct):
+            name = EvenPaddedCString()
+            data = JsonCString(length = AUTOSIZED)
+        
+        self.assertEqual(7, Person.sizeof(Person("Eli", {})))
+
+class NonCPythonTests(TestCase):
+    def setUp(self):
+        if hasattr(sys, "_getframe"):
+            self._getframe = sys._getframe
+            del sys._getframe
+    
+    def tearDown(self):
+        sys._getframe = self._getframe
+    
+    def test_parser(self):
+        self.assertRaises(CError, Parser)
+        Parser(module = __import__(__name__))
+    
+    def test_handler(self):
+        class Handler(ProtHandler): pass
+        self.assertRaises(CError, Handler)
+        
+        class Handler(ProtHandler):
+            STRUCT_MOD = __import__(__name__)
+        Handler()
 
 class UnderscorizeTests(TestCase):
     def test_camelcase(self):
@@ -764,12 +898,12 @@ class ServerTestBase:
         self.reset_logs()
     
     def send(self, data):
-        self.f.write( str(data) )
+        self.f.write(data.serialize() if isinstance(data, CStruct) else data)
         self.f.flush()
     
     def client_setup(self):
         self.sock = socket.create_connection(SERVER_ADDR)
-        self.f = self.sock.makefile("r+b", bufsize=0)
+        self.f = self.sock.makefile("rwb", 0)
     
     def client_teardown(self):
         self.f.close()
@@ -781,8 +915,6 @@ class ServerTestBase:
             if handlers and handlers[0].stream:
                 handlers[0].stream.truncate(0)
     
-    def reset_logs(self): pass
-    
     def read_log(self, name):
         with open("unit_tests." + name + "_log") as f:
             return f.read()
@@ -796,7 +928,7 @@ class ServerTests(ServerTestBase):
     
     def test_string_response(self):
         self.send(self.rp)
-        self.assertEqual(self.f.readline(), "Hello World!\n")
+        self.assertEqual(self.f.readline(), b"Hello World!\n")
     
     def test_too_short(self):
         self.send( self.np.serialize()[:5] )
@@ -809,7 +941,7 @@ class ServerTests(ServerTestBase):
         self.assertTrue("handler not defined" in self.read_log("error"))
     
     def test_unknown(self):
-        self.send("raw data")
+        self.send(b"raw data")
         sleep(0.2)
         self.assertTrue("unable to resolve" in self.read_log("error"))
     
@@ -831,6 +963,11 @@ class ServerTests(ServerTestBase):
         self.send( pg.serialize()[:-1] )
         sleep(0.2)
         self.assertTrue("struct received only" in self.read_log("error"))
+    
+    def test_autosized_too_short(self):
+        self.send(Person(name = "Eli").serialize()[:-1])
+        sleep(0.2)
+        self.assertTrue("struct received only" in self.read_log("error"))
 
 class TCPServerTests(ServerTests, TestCase):
     ServerClass = LoggingTCPServer
@@ -847,7 +984,7 @@ class UDPServerTests(ServerTests, TestCase):
         self.f = self.sock.makefile("rb")
     
     def send(self, data):
-        self.sock.sendto(str(data), SERVER_ADDR)
+        self.sock.sendto(data.serialize() if isinstance(data, CStruct) else data, SERVER_ADDR)
 
 class TCPTimeoutTests(ServerTestBase, TestCase):
     ServerClass = LoggingTCPServer
@@ -855,6 +992,7 @@ class TCPTimeoutTests(ServerTestBase, TestCase):
         timeout = 2
     
     def test_immediate_timeout(self):
+        sleep(0.1)
         before = time()
         self.server.server_close()
         self.server.shutdown()
@@ -869,7 +1007,7 @@ class TCPReadRecvInteract(ServerTestBase, TestCase):
     ServerClass = LoggingTCPServer
     class HandlerClass(TCPHandler):
         def named_origin(self, no):
-            NamedOrigin.parse(self.rfile)
+            self.log_struct( NamedOrigin.parse(self.rfile) )
             return NamedOrigin()
     
     def test_recv_read_interaction(self):
@@ -879,7 +1017,7 @@ class TCPReadRecvInteract(ServerTestBase, TestCase):
             NamedOrigin.parse(self.f)
 
 if __name__ == "__main__":
-    logging.basicConfig(level = logging.INFO)
+    logging.basicConfig(level = logging.DEBUG)
     delete_logs()
     atexit.register(delete_logs)
     main()
